# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Polynomiality rules."""
from suspect.polynomial.degree import PolynomialDegree
from suspect.expression import ExpressionType
from suspect.interfaces import Rule

__all__ = [
    'VariableRule', 'ConstantRule', 'ConstraintRule', 'ObjectiveRule', 'DivisionRule',
    'ProductRule', 'LinearRule', 'SumRule', 'NegationRule', 'PowerRule', 'UnaryFunctionRule',
    'QuadraticRule',
]


class VariableRule(Rule):
    """Return polynomial degree of variable."""
    root_expr = ExpressionType.Variable

    def apply(self, _expr, _ctx):
        return PolynomialDegree(1)


class ConstantRule(Rule):
    """Return polynomial degree of constant."""
    root_expr = ExpressionType.Constant

    def apply(self, _expr, _ctx):
        return PolynomialDegree(0)


class ConstraintRule(Rule):
    """Return polynomial degree of constraint."""
    root_expr = ExpressionType.Constraint

    def apply(self, expr, ctx):
        return ctx[expr.children[0]]


class ObjectiveRule(Rule):
    """Return polynomial degree of objective."""
    root_expr = ExpressionType.Objective

    def apply(self, expr, ctx):
        return ctx[expr.children[0]]


class DivisionRule(Rule):
    """Return polynomial degree of division.

    If the denominator is constant, it returns the polynomial degree of the numerator,
    otherwise return no polynomial degree.
    """
    root_expr = ExpressionType.Division

    def apply(self, expr, ctx):
        den_degree = ctx[expr.children[1]]
        if den_degree.is_polynomial() and den_degree.degree == 0:
            # constant denominator
            return ctx[expr.children[0]]
        return PolynomialDegree.not_polynomial()


class ProductRule(Rule):
    """Return polynomial degree of product.

    The polynomial degree of a product is equal to the sum of the polynomial degre of its children.
    """
    root_expr = ExpressionType.Product

    def apply(self, expr, ctx):
        return sum([ctx[a] for a in expr.children], PolynomialDegree(0))


class LinearRule(Rule):
    """Return polynomial degree of a linear expression."""
    root_expr = ExpressionType.Linear

    def apply(self, expr, ctx):
        if not expr.children:
            return PolynomialDegree(0)
        return PolynomialDegree(1)


class QuadraticRule(Rule):
    """Return polynomial degree of a quadratic expression."""
    root_expr = ExpressionType.Quadratic

    def apply(self, expr, ctx):
        if not expr.children:
            return PolynomialDegree(0)
        return PolynomialDegree(2)


class SumRule(Rule):
    """Return polynomial degree of a summation expression."""
    root_expr = ExpressionType.Sum

    def apply(self, expr, ctx):
        return max(ctx[a] for a in expr.children)


class NegationRule(Rule):
    """Return polynomial degree of a negation expression."""
    root_expr = ExpressionType.Negation

    def apply(self, expr, ctx):
        return ctx[expr.children[0]]


class PowerRule(Rule):
    """Return polynomial degree of a power expression."""
    root_expr = ExpressionType.Power

    def apply(self, expr, ctx):
        assert len(expr.children) == 2
        base, expo = expr.children
        base_degree = ctx[base]
        expo_degree = ctx[expo]

        if not (base_degree.is_polynomial() and expo_degree.is_polynomial()):
            return PolynomialDegree.not_polynomial()

        if expo_degree.degree == 0:
            if base_degree.degree == 0:
                # const ** const
                return PolynomialDegree(0)
            if not expo.is_constant():
                return PolynomialDegree.not_polynomial()
            expo = expo.value
            if expo == int(expo):
                return base_degree ** expo
        return PolynomialDegree.not_polynomial()


class UnaryFunctionRule(Rule):
    """Return polynomial degree of an unary function."""
    root_expr = ExpressionType.UnaryFunction

    def apply(self, expr, ctx):
        return PolynomialDegree.not_polynomial()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for negation."""
from suspect.expression import ExpressionType
from suspect.interfaces import Rule


class NegationRule(Rule):
    """Return convexity of negation."""
    root_expr = ExpressionType.Negation

    def apply(self, expr, ctx):
        child = expr.children[0]
        cvx = ctx.convexity(child)
        return cvx.negate()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pkg_resources
from suspect.monotonicity import MonotonicityPropagationVisitor
from suspect.convexity import ConvexityPropagationVisitor
from suspect.dag.iterator import DagForwardIterator


logger = logging.getLogger('suspect')


def monotonicity_detection_entry_points():
    return pkg_resources.iter_entry_points('suspect.monotonicity_detection')


def convexity_detection_entry_points():
    return pkg_resources.iter_entry_points('suspect.convexity_detection')


class SpecialStructurePropagationVisitor(object):
    def __init__(self, problem):
        self._mono_visitors = [MonotonicityPropagationVisitor()]
        for entry_point in monotonicity_detection_entry_points():
            cls = entry_point.load()
            self._mono_visitors.append(cls(problem))

        logger.info(
            'Loaded %s monotonicity detectors: %s',
            len(self._mono_visitors), ', '.join([str(mono) for mono in self._mono_visitors])
        )

        self._cvx_visitors = [ConvexityPropagationVisitor()]
        for entry_point in convexity_detection_entry_points():
            cls = entry_point.load()
            self._cvx_visitors.append(cls(problem))

        logger.info(
            'Loaded %s convexity detectors: %s',
            len(self._cvx_visitors), ', '.join([str(cvx) for cvx in self._cvx_visitors])
        )

    def visit(self, expr, ctx):
        for mono_visitor in self._mono_visitors:
            mono_known = mono_visitor.visit(expr, ctx)
            if mono_known:
                break

        for cvx_visitor in self._cvx_visitors:
            cvx_known = cvx_visitor.visit(expr, ctx)
            if cvx_known:
                break
        return [expr]


def propagate_special_structure(problem, ctx):
    """Propagate special structure.

    Arguments
    ---------
    problem: ProblemDag
        the problem.
    ctx: SpecialStructurePropagationContext
      the context containing the bounds

    Returns
    -------
    monotonicity: dict-like
        monotonicity information for the problem.
    convexity: dict-like
        convexity information for the problem.
    """
    visitor = SpecialStructurePropagationVisitor(problem)
    iterator = DagForwardIterator()
    iterator.iterate(problem, visitor, ctx)
    return ctx.monotonicity, ctx.convexity



# pylint: skip-file
import pytest
from tests.conftest import PlaceholderExpression
import suspect.dag.expressions as dex
from suspect.interval import Interval
from suspect.fbbt.initialization import BoundsInitializationVisitor


class InitContext:
    def __init__(self):
        self._bounds = {}

    def set_bounds(self, expr, value):
        self._bounds[expr] = value

    def bounds(self,  expr):
        return self._bounds[expr]


@pytest.fixture
def visitor():
    return BoundsInitializationVisitor()


def _visit(visitor, root_cls):
    p = PlaceholderExpression()
    root = root_cls([p])
    ctx = InitContext()
    visitor.visit(root, ctx)
    return ctx.bounds(p)


def test_sqrt(visitor):
    assert _visit(visitor, dex.SqrtExpression).is_nonnegative()


def test_log(visitor):
    assert _visit(visitor, dex.LogExpression).is_nonnegative()


def test_asin(visitor):
    assert _visit(visitor, dex.AsinExpression) == Interval(-1, 1)


def test_acos(visitor):
    assert _visit(visitor, dex.AcosExpression) == Interval(-1, 1)



# pylint: skip-file
import pytest
from hypothesis import assume, given, reproduce_failure
import hypothesis.strategies as st
import itertools
from tests.conftest import coefficients, reals
from tests.conftest import PlaceholderExpression as PE
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.monotonicity import Monotonicity as M
from suspect.convexity import Convexity as C
from suspect.interval import Interval as I
from suspect.convexity.rules import *
from suspect.math import almosteq, pi


class ConvexityContext:
    def __init__(self, cvx=None, mono=None, bounds=None):
        if cvx is None:
            cvx = {}
        if mono is None:
            mono = {}
        if bounds is None:
            bounds = {}
        self._cvx = cvx
        self._mono = mono
        self._bounds = bounds

    def convexity(self, expr):
        return self._cvx[expr]

    def set_convexity(self, expr, value):
        self._cvx[expr] = value

    def monotonicity(self, expr):
        return self._mono[expr]

    def bounds(self, expr):
        return self._bounds[expr]


@st.composite
def convex_terms(draw):
    coef = draw(coefficients())
    if coef > 0:
        return (coef, C.Convex)
    else:
        return (coef, C.Concave)


@st.composite
def concave_terms(draw):
    coef = draw(coefficients())
    if coef < 0:
        return (coef, C.Convex)
    else:
        return (coef, C.Concave)


@st.composite
def linear_terms(draw):
    coef = draw(coefficients())
    return (coef, C.Linear)


def test_variable_is_linear():
    rule = VariableRule()
    ctx = ConvexityContext()
    result = rule.checked_apply(PE(ET.Variable), ctx)
    assert result.is_linear()


def test_constant_is_linear():
    rule = ConstantRule()
    ctx = ConvexityContext()
    result = rule.checked_apply(PE(ET.Constant), ctx)
    assert result.is_linear()


@pytest.mark.parametrize('cvx_child,bounded_below,bounded_above,expected', [
    # l <= g(x) <= u is always unknown except if g(x) is linear
    (C.Linear, True, True, C.Linear),
    (C.Convex, True, True, C.Unknown),
    (C.Concave, True, True, C.Unknown),
    (C.Unknown, True, True, C.Unknown),

    # g(x) <= u is the same as g(x)
    (C.Linear, False, True, C.Linear),
    (C.Convex, False, True, C.Convex),
    (C.Concave, False, True, C.Concave),
    (C.Unknown, False, True, C.Unknown),

    # l <= g(x) is the negation of g(x)
    (C.Linear, True, False, C.Linear),
    (C.Convex, True, False, C.Concave),
    (C.Concave, True, False, C.Convex),
    (C.Unknown, True, False, C.Unknown),
])
def test_constraint(cvx_child, bounded_below, bounded_above, expected):
    rule = ConstraintRule()
    child = PE()
    ctx = ConvexityContext({child: cvx_child})
    result = rule.checked_apply(
        PE(ET.Constraint, [child], bounded_below=bounded_below, bounded_above=bounded_above),
        ctx,
    )
    assert result == expected


@pytest.mark.parametrize('cvx_child,expected', [
    (C.Linear, C.Linear), (C.Convex, C.Convex), (C.Concave, C.Concave), (C.Unknown, C.Unknown),
])
def test_objective_is_minimizing(cvx_child, expected):
    rule = ObjectiveRule()
    child = PE()
    ctx = ConvexityContext({child: cvx_child})
    result = rule.checked_apply(
        PE(ET.Objective, [child], is_minimizing=True),
        ctx,
    )
    assert result == expected


@pytest.mark.parametrize('cvx_child,expected', [
    (C.Linear, C.Linear), (C.Convex, C.Concave), (C.Concave, C.Convex), (C.Unknown, C.Unknown),
])
def test_objective_is_maximizing(cvx_child, expected):
    rule = ObjectiveRule()
    child = PE()
    ctx = ConvexityContext({child: cvx_child})
    result = rule.checked_apply(
        PE(ET.Objective, [child], is_minimizing=False),
        ctx,
    )
    assert result == expected


@pytest.mark.parametrize('cvx,expected', [
    (C.Convex, C.Concave), (C.Concave, C.Convex), (C.Linear, C.Linear), (C.Unknown, C.Unknown),
])
def test_negation(cvx, expected):
    rule = NegationRule()
    child = PE()
    ctx = ConvexityContext(cvx={child: cvx})
    result = rule.checked_apply(PE(ET.Negation, [child]), ctx)
    assert result == expected


class TestProduct:
    def _rule_result(self, cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g):
        rule = ProductRule()
        f = PE()
        g = PE()
        ctx = ConvexityContext(
            cvx={f: cvx_f, g: cvx_g},
            mono={f: mono_f, g: mono_g},
            bounds={f: bounds_f, g: bounds_g},
        )
        return rule.checked_apply(PE(ET.Product, [f, g]), ctx)

    @pytest.mark.parametrize('cvx_f,bounds_g,expected', [
        (C.Convex, I(0, None), C.Convex),
        (C.Convex, I(None, 0), C.Concave),
        (C.Concave, I(0, None), C.Concave),
        (C.Concave, I(None, 0), C.Convex),
        (C.Linear, I(0, None), C.Linear),
        (C.Linear, I(None, 0), C.Linear),
        (C.Unknown, I(0, None), C.Unknown),
        (C.Unknown, I(None, 0), C.Unknown),
    ])
    def test_product_with_constant(self, cvx_f, bounds_g, expected):
        cvx_g = C.Linear # g is constant
        mono_f = M.Unknown
        mono_g = M.Constant
        bounds_f = I(None, None)
        assert self._rule_result(cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g) == expected
        assert self._rule_result(cvx_g, cvx_f, mono_g, mono_f, bounds_g, bounds_f) == expected

    @pytest.mark.parametrize('cvx_f,bounds_f,expected', [
        (C.Linear, I(None, None), C.Convex),
        (C.Linear, I(0, None), C.Convex),
        (C.Linear, I(None, 0), C.Convex),

        (C.Convex, I(None, None), C.Unknown),
        (C.Convex, I(0, None), C.Convex),
        (C.Convex, I(None, 0), C.Unknown),

        (C.Concave, I(None, None), C.Unknown),
        (C.Concave, I(0, None), C.Unknown),
        (C.Concave, I(None, 0), C.Convex),
    ])
    def test_product_with_itself(self, cvx_f, bounds_f, expected):
        rule = ProductRule()
        f = PE()
        ctx = ConvexityContext(
            cvx={f: cvx_f},
            bounds={f: bounds_f},
        )
        assert rule.checked_apply(PE(ET.Product, [f, f]), ctx) == expected

    @given(reals())
    def test_product_with_itself_with_coeff(self, coef):
        if coef > 0:
            expected = C.Convex
        else:
            expected = C.Concave
        rule = ProductRule()
        v = PE(ET.Variable)
        l = PE(ET.Linear, [v], coefficients=[coef])
        ctx = ConvexityContext()
        assert rule.checked_apply(PE(ET.Product, [l, v]), ctx) == expected
        assert rule.checked_apply(PE(ET.Product, [v, l]), ctx) == expected

    def test_product_linear_by_var(self):
        rule = ProductRule()
        v = PE(ET.Variable)
        l = PE(ET.Linear, [PE(ET.Variable), PE(ET.Variable)], coefficients=[1.0, 2.0])
        ctx = ConvexityContext(mono={v: M.Nondecreasing, l: M.Nondecreasing})
        assert rule.checked_apply(PE(ET.Product, [l, v]), ctx) == C.Unknown
        assert rule.checked_apply(PE(ET.Product, [v, l]), ctx) == C.Unknown


class TestDivision:
    def _rule_result(self, cvx_f, cvx_g, mono_f, mono_g, bounds_f, bounds_g):
        rule = DivisionRule()
        f = PE()
        g = PE()
        ctx = ConvexityContext(
            cvx={f: cvx_f, g: cvx_g},
            mono={f: mono_f, g: mono_g},
            bounds={f: bounds_f, g: bounds_g},
        )
        return rule.checked_apply(PE(ET.Division, [f, g]), ctx)

    @pytest.mark.parametrize('cvx_f,bounds_g,expected', [
        (C.Convex, I(1e-5, None), C.Convex),
        (C.Convex, I(0, None), C.Unknown),

        (C.Concave, I(None, -1e-5), C.Convex),
        (C.Concave, I(None, 0), C.Unknown),

        (C.Convex, I(None, -1e-5), C.Concave),
        (C.Convex, I(None, 0), C.Unknown),

        (C.Concave, I(1e-5, None), C.Concave),
        (C.Concave, I(0, None), C.Unknown),
    ])
    def test_constant_denominator(self, cvx_f, bounds_g, expected):
        assert self._rule_result(cvx_f, C.Linear, M.Unknown, M.Constant, I(None, None), bounds_g) == expected


    @pytest.mark.parametrize('cvx_g,bounds_f,bounds_g,expected', [
        (C.Concave, I(0, None), I(1e-5, None), C.Convex),
        (C.Concave, I(0, None), I(0, None), C.Unknown),
        (C.Convex, I(None, 0), I(None, -1e-5), C.Convex),
        (C.Convex, I(None, 0), I(None, 0), C.Unknown),
        (C.Convex, I(0, None), I(None, -1e-5), C.Concave),
        (C.Convex, I(0, None), I(None, 0), C.Unknown),
        (C.Concave, I(None, 0), I(1e-5, None), C.Concave),
        (C.Concave, I(None, 0), I(0, None), C.Unknown),
    ])
    def test_constant_numerator(self, cvx_g, bounds_f, bounds_g, expected):
        assert self._rule_result(C.Linear, cvx_g, M.Constant, M.Unknown, bounds_f, bounds_g) == expected


class TestLinear:
    def _rule_result(self, terms):
        coefficients = [c for c, _ in terms]
        children = [PE() for _ in terms]
        cvxs = [c for _, c in terms]
        ctx = ConvexityContext()
        for child, cvx in zip(children, cvxs):
            ctx.set_convexity(child, cvx)
        rule = LinearRule()
        return rule.checked_apply(PE(ET.Linear, children=children, coefficients=coefficients), ctx)

    @given(
        st.lists(convex_terms(), min_size=1),
        st.lists(linear_terms()),
    )
    def test_convex(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_convex()

    @given(
        st.lists(concave_terms(), min_size=1),
        st.lists(linear_terms()),
    )
    def test_concave(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_concave()

    @given(
        st.lists(linear_terms(), min_size=1),
    )
    def test_linear(self, a):
        cvx = self._rule_result(a)
        assert cvx.is_linear()

    @given(
        st.lists(convex_terms(), min_size=1),
        st.lists(concave_terms(), min_size=1),
    )
    def test_unknown(self, a, b):
        terms = a + b
        for c, _ in terms:
            assume(c != 0.0)
        cvx = self._rule_result(terms)
        assert cvx.is_unknown()


class TestSum:
    def _rule_result(self, cvxs):
        children = [PE() for _ in cvxs]
        ctx = ConvexityContext()
        for child, cvx in zip(children, cvxs):
            ctx.set_convexity(child, cvx)
        rule = SumRule()
        return rule.checked_apply(PE(ET.Sum, children=children), ctx)

    @given(
        st.lists(st.just(C.Convex), min_size=1),
        st.lists(st.just(C.Linear)),
    )
    def test_convex(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_convex()

    @given(
        st.lists(st.just(C.Concave), min_size=1),
        st.lists(st.just(C.Linear)),
    )
    def test_concave(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_concave()

    @given(
        st.lists(st.just(C.Linear), min_size=1),
    )
    def test_linear(self, a):
        cvx = self._rule_result(a)
        assert cvx.is_linear()

    @given(
        st.lists(st.just(C.Concave), min_size=1),
        st.lists(st.just(C.Convex), min_size=1),
    )
    def test_unknown(self, a, b):
        cvx = self._rule_result(a + b)
        assert cvx.is_unknown()


class UnaryFunctionTest:
    rule_cls = None
    func_type = None

    def _rule_result(self, cvx, mono, bounds):
        rule = self.rule_cls()
        child = PE()
        ctx = ConvexityContext({child: cvx}, {child: mono}, {child: bounds})
        return rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=self.func_type), ctx)

class TestAbs(UnaryFunctionTest):
    rule_cls = AbsRule
    func_type = UFT.Abs

    @pytest.mark.parametrize('mono,bounds',
                             itertools.product([M.Nondecreasing, M.Nonincreasing, M.Constant],
                                               [I(None, 0), I(0, None), I(None, None)]))
    def test_linear_child(self, mono, bounds):
        cvx = self._rule_result(C.Linear, mono, bounds)
        assert cvx == C.Convex

    @pytest.mark.parametrize('mono,bounds,expected', [
        (M.Unknown, I(0, None), C.Convex),
        (M.Unknown, I(None, 0), C.Concave),
    ])
    def test_convex_child(self, mono, bounds, expected):
        cvx = self._rule_result(C.Convex, mono, bounds)
        assert cvx == expected

    @pytest.mark.parametrize('mono,bounds,expected', [
        (M.Unknown, I(0, None), C.Concave),
        (M.Unknown, I(None, 0), C.Convex),
    ])
    def test_concave_child(self, mono, bounds, expected):
        cvx = self._rule_result(C.Concave, mono, bounds)
        assert cvx == expected




class TestSqrt(UnaryFunctionTest):
    rule_cls = SqrtRule
    func_type = UFT.Sqrt

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Concave, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        )
    )
    def test_concave_child(self, cvx, mono):
        cvx = self._rule_result(cvx, mono, I(0, None))
        assert cvx == C.Concave

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Convex, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        ),
    )
    def test_non_concave_child(self, cvx, mono):
        cvx = self._rule_result(cvx, mono, I(0, None))
        assert cvx == C.Unknown


class TestExp(UnaryFunctionTest):
    rule_cls = ExpRule
    func_type = UFT.Exp

    @pytest.mark.parametrize(
        'cvx,mono,bounds',
        itertools.product(
            [C.Convex, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
            [I(0, None), I(None, 0), I(None, None)],
        )
    )
    def test_convex_child(self, cvx, mono, bounds):
        cvx = self._rule_result(cvx, mono, bounds)
        assert cvx == C.Convex

    @pytest.mark.parametrize(
        'cvx,mono,bounds',
        itertools.product(
            [C.Concave, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
            [I(0, None), I(None, 0), I(None, None)],
        )
    )
    def test_convex_child(self, cvx, mono, bounds):
        cvx = self._rule_result(cvx, mono, bounds)
        assert cvx == C.Unknown


class TestLog(UnaryFunctionTest):
    rule_cls = LogRule
    func_type = UFT.Log

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Concave, C.Linear],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        )
    )
    def test_concave_child(self, cvx, mono):
        cvx = self._rule_result(cvx, mono, I(0, None))
        assert cvx == C.Concave

    @pytest.mark.parametrize(
        'cvx,mono',
        itertools.product(
            [C.Convex, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
        ),
    )
    def test_non_concave_child(self, cvx, mono):
        cvx = self._rule_result(cvx, mono, I(0, None))
        assert cvx == C.Unknown


class TestSin(UnaryFunctionTest):
    rule_cls = SinRule
    func_type = UFT.Sin

    def test_bound_size_too_big(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(-2, 2))
        assert cvx == C.Unknown

    def test_bound_opposite_sign(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(-0.1, 0.1))
        assert cvx == C.Unknown

    def test_positive_sin_linear_child(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(pi/2-0.1, pi/2+0.1))
        assert cvx == C.Concave

    def test_positive_sin_convex_child_but_wrong_interval(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(pi/2-0.1, pi/2+0.1))
        assert cvx == C.Unknown

    def test_positive_sin_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(pi/2+0.1, pi-0.1))
        assert cvx == C.Concave

    def test_positive_sin_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(0.1, pi/2-0.1))
        assert cvx == C.Concave

    def test_negative_sin_linear_child(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(1.5*pi-0.1, 1.5*pi+0.1))
        assert cvx == C.Convex

    def test_negative_sin_concave_child_but_wrong_interval(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(1.5*pi-0.1, 1.5*pi+0.1))
        assert cvx == C.Unknown

    def test_negative_sin_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Constant, I(pi+0.1, 1.5*pi-0.1))
        assert cvx == C.Convex

    def test_negative_sin_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(1.5*pi, 2*pi))
        assert cvx == C.Convex


class TestCos(UnaryFunctionTest):
    rule_cls = CosRule
    func_type = UFT.Cos

    def test_bound_size_too_big(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(-2, 2))
        assert cvx == C.Unknown

    def test_bound_opposite_sign(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(pi/2-0.1, pi/2+0.1))
        assert cvx == C.Unknown

    def test_positive_cos_linear_child(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(0, 0.5*pi))
        assert cvx == C.Concave

    def test_positive_cos_convex_child_but_wrong_interval(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(0, 0.5*pi))
        assert cvx == C.Unknown

    def test_positive_cos_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(1.5*pi, 2*pi))
        assert cvx == C.Concave

    def test_positive_cos_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(0, 0.5*pi))
        assert cvx == C.Concave

    def test_negative_cos_linear_child(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(0.6*pi, 1.4*pi))
        assert cvx == C.Convex

    def test_negative_cos_concave_child_but_wrong_interval(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(0.6*pi, 0.9*pi))
        assert cvx == C.Unknown

    def test_negative_cos_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(1.1*pi, 1.4*pi))
        assert cvx == C.Convex

    def test_negative_cos_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(0.6*pi, 0.9*pi))
        assert cvx == C.Convex



class TestTan(UnaryFunctionTest):
    rule_cls = TanRule
    func_type = UFT.Tan

    def test_bound_size_too_big(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(-2, 2))
        assert cvx == C.Unknown

    def test_bound_opposite_sign(self):
        cvx = self._rule_result(C.Linear, M.Constant, I(-0.1, 0.1))
        assert cvx == C.Unknown

    def test_positive_tan_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(pi, 1.5*pi))
        assert cvx == C.Convex

    def test_positive_tan_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(pi, 1.5*pi))
        assert cvx == C.Unknown

    def test_negative_tan_convex_child(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(-1.5*pi, -pi))
        assert cvx == C.Unknown

    def test_negative_tan_concave_child(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(-0.49*pi, -0.1))
        assert cvx == C.Concave


class TestAsin(UnaryFunctionTest):
    rule_cls = AsinRule
    func_type = UFT.Asin

    def test_is_concave(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(-1, 0))
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(-1, 0.1), I(-1.1, 0.0), I(None, None)])
    def test_is_not_concave(self, bounds):
        cvx = self._rule_result(C.Concave, M.Unknown, bounds)
        assert cvx == C.Unknown

    def test_is_convex(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(0, 1))
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(-0.1, 1), I(0.0, 1.1), I(None, None)])
    def test_is_not_convex(self, bounds):
        cvx = self._rule_result(C.Convex, M.Unknown, bounds)
        assert cvx == C.Unknown


class TestAcos(UnaryFunctionTest):
    rule_cls = AcosRule
    func_type = UFT.Acos

    def test_is_concave(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(0, 1))
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(-0.1, 1), I(0.0, 1.1), I(None, None)])
    def test_is_not_concave(self, bounds):
        cvx = self._rule_result(C.Convex, M.Unknown, bounds)
        assert cvx == C.Unknown

    def test_is_convex(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(-1, 0))
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(-1, 0.1), I(-1.1, 0.0), I(None, None)])
    def test_is_not_convex(self, bounds):
        cvx = self._rule_result(C.Concave, M.Unknown, bounds)
        assert cvx == C.Unknown


class TestAtan(UnaryFunctionTest):
    rule_cls = AtanRule
    func_type = UFT.Atan

    def test_is_concave(self):
        cvx = self._rule_result(C.Concave, M.Unknown, I(0, None))
        assert cvx == C.Concave

    @pytest.mark.parametrize('bounds', [I(None, 0), I(None, None)])
    def test_is_not_concave(self, bounds):
        cvx = self._rule_result(C.Concave, M.Unknown, bounds)
        assert cvx == C.Unknown

    def test_is_convex(self):
        cvx = self._rule_result(C.Convex, M.Unknown, I(None, 0))
        assert cvx == C.Convex

    @pytest.mark.parametrize('bounds', [I(0, None), I(None, None)])
    def test_is_not_convex(self, bounds):
        cvx = self._rule_result(C.Convex, C.Unknown, bounds)
        assert cvx == C.Unknown


class TestPowConstantBase:
    def _rule_result(self, base, cvx_expo, mono_expo, bounds_expo):
        base = PE(ET.Constant, value=base)
        expo = PE()
        ctx = ConvexityContext(
            cvx={expo: cvx_expo, base: C.Linear},
            mono={expo: mono_expo, base: M.Constant},
            bounds={base: I(base.value, base.value), expo: bounds_expo},
        )
        rule = PowerRule()
        return rule.checked_apply(PE(ET.Power, [base, expo]), ctx)

    @pytest.mark.parametrize(
        'cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(base=reals(max_value=-0.01, allow_infinity=False))
    def test_negative_base(self, base, cvx_expo, mono_expo, bounds_expo):
        cvx = self._rule_result(base, cvx_expo, mono_expo, bounds_expo)
        assert cvx == C.Unknown

    @pytest.mark.parametrize('cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(base=reals(min_value=0.001, max_value=0.999))
    def test_base_between_0_and_1(self, base, cvx_expo, mono_expo, bounds_expo):
        if cvx_expo == C.Concave or cvx_expo == C.Linear:
            expected = C.Convex
        else:
            expected = C.Unknown
        cvx = self._rule_result(base, cvx_expo, mono_expo, bounds_expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_expo,mono_expo,bounds_expo',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    @given(base=reals(min_value=1, allow_infinity=False))
    def test_base_gt_1(self, base, cvx_expo, mono_expo, bounds_expo):
        if cvx_expo == C.Convex or cvx_expo == C.Linear:
            expected = C.Convex
        else:
            expected = C.Unknown
        cvx = self._rule_result(base, cvx_expo, mono_expo, bounds_expo)
        assert cvx == expected


class TestPowConstantExponent(object):
    def _rule_result(self, cvx_base, mono_base, bounds_base, expo):
        expo = PE(ET.Constant, value=expo)
        base = PE()
        ctx = ConvexityContext(
            cvx={base: cvx_base, expo: C.Linear},
            mono={base: mono_base, expo: M.Constant},
            bounds={base: bounds_base, expo: I(expo.value, expo.value)},
        )
        rule = PowerRule()
        return rule.checked_apply(PE(ET.Power, [base, expo]), ctx)

    @pytest.mark.parametrize(
        'cvx_base,mono_base,bounds_base',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    def test_exponent_equals_0(self, cvx_base, mono_base, bounds_base):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, 0.0)
        assert cvx == C.Linear

    @pytest.mark.parametrize(
        'cvx_base,mono_base,bounds_base',
        itertools.product(
            [C.Convex, C.Concave, C.Linear, C.Unknown],
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)],
        )
    )
    def test_exponent_equals_1(self, cvx_base, mono_base, bounds_base):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, 1.0)
        assert cvx == cvx_base

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Linear, M.Nondecreasing, I(None, None), C.Convex),

        (C.Convex, M.Unknown, I(0, None), C.Convex),
        (C.Convex, M.Unknown, I(None, 0), C.Unknown),

        (C.Concave, M.Unknown, I(0, None), C.Unknown),
        (C.Concave, M.Unknown, I(None, 0), C.Convex),
    ])
    @given(expo=st.integers(min_value=1))
    def test_positive_even_integer(self, expo, cvx_base, mono_base, bounds_base, expected):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, 2*expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Convex, M.Unknown, I(None, 0), C.Convex),
        (C.Convex, M.Unknown, I(0, None), C.Concave),

        (C.Concave, M.Unknown, I(0, None), C.Convex),
        (C.Concave, M.Unknown, I(None, 0), C.Concave),
    ])
    @given(expo=st.integers(min_value=1))
    def test_negative_even_integer(self, expo, cvx_base, mono_base, bounds_base, expected):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, -2*expo)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Convex, M.Unknown, I(0, None), C.Convex),
        (C.Convex, M.Unknown, I(None, 0), C.Unknown),
        (C.Concave, M.Unknown, I(None, 0), C.Concave),
        (C.Concave, M.Unknown, I(0, None), C.Unknown),
    ])
    @given(expo=st.integers(min_value=1))
    def test_positive_odd_integer(self, expo, cvx_base, mono_base, bounds_base, expected):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, 2*expo+1)
        assert cvx == expected

    @pytest.mark.parametrize('cvx_base,mono_base,bounds_base,expected', [
        (C.Concave, M.Unknown, I(0, None), C.Convex),
        (C.Concave, M.Unknown, I(None, 0), C.Unknown),
        (C.Convex, M.Unknown, I(None, 0), C.Concave),
        (C.Convex, M.Unknown, I(0, None), C.Unknown),
    ])
    @given(expo=st.integers(min_value=1))
    def test_negative_odd_integer(self, expo, cvx_base, mono_base, bounds_base, expected):
        cvx = self._rule_result(cvx_base, mono_base, bounds_base, -2*expo+1)
        assert cvx == expected

    @given(expo=reals(min_value=1, allow_infinity=False))
    def test_positive_gt_1_non_integer_negative_base(self, expo):
        expo = expo + 1e-6
        assume(expo != int(expo))
        cvx = self._rule_result(C.Convex, M.Unknown, I(None, -1), expo)
        assert cvx == C.Unknown

    @given(expo=reals(min_value=1, allow_infinity=False))
    def test_positive_gt_1_non_integer(self, expo):
        expo = expo + 1e-5 # make it positive
        assume(expo != int(expo))
        cvx = self._rule_result(C.Convex, M.Unknown, I(0, None), expo)
        assert cvx == C.Convex

    @pytest.mark.parametrize('cvx,expected', [(C.Convex, C.Concave), (C.Concave, C.Convex)])
    @given(expo=reals(max_value=0, allow_infinity=False))
    def test_positive_lt_0_non_integer(self, expo, cvx, expected):
        expo = expo - 1e-5 # make it negative
        assume(not almosteq(expo, int(expo)))
        cvx = self._rule_result(cvx, M.Unknown, I(0, None), expo)
        assert cvx == expected

    @given(expo=reals(min_value=0, max_value=1, allow_infinity=False))
    def test_positive_0_1_non_integer(self, expo):
        assume(not almosteq(expo, int(expo)))
        cvx = self._rule_result(C.Concave, M.Unknown, I(0, None), expo)
        assert cvx == C.Concave



# pylint: skip-file
import pytest
from hypothesis import given
import hypothesis.strategies as st
from tests.conftest import PlaceholderExpression as PE
from suspect.expression import ExpressionType as ET
from suspect.polynomial.degree import PolynomialDegree
from suspect.polynomial.rules import *


@st.composite
def polynomial_degrees(draw, allow_not_polynomial=True):
    if allow_not_polynomial:
        degree = draw(st.one_of(st.none(), st.integers(min_value=0)))
    else:
        degree = draw(st.integers(min_value=0))
    return PolynomialDegree(degree)


@st.composite
def placeholder_expressions(draw):
    return PE(draw(st.sampled_from(ET)))


def test_variable_is_always_1():
    rule = VariableRule()
    result = rule.checked_apply(PE(ET.Variable), None)
    assert result.degree == 1


def test_constant_is_always_0():
    rule = ConstantRule()
    result = rule.checked_apply(PE(ET.Constant), None)
    assert result.degree == 0


@given(polynomial_degrees())
@pytest.mark.parametrize('rule_cls,root_type', [
    (ConstraintRule ,ET.Constraint),
    (ObjectiveRule, ET.Objective),
    (NegationRule, ET.Negation),
])
def test_passtrough_rules(rule_cls, root_type, child_degree):
    rule = rule_cls()
    child = PE(ET.Variable)
    ctx = {child: child_degree}
    result = rule.checked_apply(PE(root_type, [child]), ctx)
    assert result == child_degree


@given(polynomial_degrees())
def test_division_rule_with_constant_denominator(child_degree):
    rule = DivisionRule()
    num = PE(ET.Variable)
    den = PE(ET.Variable)
    ctx = {
        num: child_degree,
        den: PolynomialDegree(0),
    }
    result = rule.checked_apply(PE(ET.Division, [num, den]), ctx)
    assert result == child_degree


@given(polynomial_degrees())
def test_division_rule_with_nonconstant_denominator(child_degree):
    rule = DivisionRule()
    num = PE(ET.Variable)
    den = PE(ET.Variable)
    ctx = {
        num: child_degree,
        den: PolynomialDegree(1),
    }
    result = rule.checked_apply(PE(ET.Division, [num, den]), ctx)
    assert not result.is_polynomial()


@given(st.lists(st.tuples(polynomial_degrees(), placeholder_expressions()), min_size=1))
def test_product_rule(children_degrees):
    children = []
    ctx = {}
    expected = 0
    for degree, child in children_degrees:
        children.append(child)
        ctx[child] = degree
        if expected is None:
            continue
        if degree.is_polynomial():
            expected += degree.degree
        else:
            expected = None
    expected = PolynomialDegree(expected)
    rule = ProductRule()
    result = rule.checked_apply(PE(ET.Product, children), ctx)
    assert result == expected


def test_linear_rule_with_no_children():
    rule = LinearRule()
    result = rule.checked_apply(PE(ET.Linear, []), None)
    assert result.degree == 0


@given(st.integers(min_value=1))
def test_linear_rule_with_children(children_size):
    rule = LinearRule()
    children = [PE(ET.Variable)] * children_size
    result = rule.checked_apply(PE(ET.Linear, children), None)
    assert result.degree == 1


@given(st.lists(st.tuples(polynomial_degrees(), placeholder_expressions()), min_size=1))
def test_sum_rule(children_degrees):
    children = []
    ctx = {}
    expected = 0
    for degree, child in children_degrees:
        children.append(child)
        ctx[child] = degree
        if expected is None:
            continue
        if degree.is_polynomial() and degree.degree > expected:
            expected = degree.degree
        if not degree.is_polynomial():
            expected = None

    expected = PolynomialDegree(expected)
    rule = SumRule()
    result = rule.checked_apply(PE(ET.Sum, children), ctx)
    assert result == expected


def test_power_with_non_polynomial_exponent():
    ctx = {}
    base = PE(ET.Variable)
    ctx[base] = PolynomialDegree(1)
    expo = PE(ET.UnaryFunction)
    ctx[expo] = PolynomialDegree(None)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert not result.is_polynomial()


def test_power_with_non_polynomial_base():
    ctx = {}
    base = PE(ET.UnaryFunction)
    ctx[base] = PolynomialDegree(None)
    expo = PE(ET.Variable)
    ctx[expo] = PolynomialDegree(1)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert not result.is_polynomial()


def test_power_constant_power_constant():
    ctx = {}
    base = PE(ET.Constant)
    ctx[base] = PolynomialDegree(0)
    expo = PE(ET.Constant)
    ctx[expo] = PolynomialDegree(0)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert result.is_polynomial() and result.degree == 0


def test_power_non_constant():
    ctx = {}
    base = PE(ET.Variable)
    ctx[base] = PolynomialDegree(1)
    expo = PE(ET.Constant)
    ctx[expo] = PolynomialDegree(0)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert not result.is_polynomial()


def test_power_non_constant_polynomial():
    ctx = {}
    base = PE(ET.Variable)
    ctx[base] = PolynomialDegree(1)
    expo = PE(ET.Product)
    ctx[expo] = PolynomialDegree(4)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert not result.is_polynomial()


def test_power_constant_value():
    ctx = {}
    base = PE(ET.Product)
    ctx[base] = PolynomialDegree(2)
    expo = PE(ET.Constant, is_constant=True, value=4.0)
    ctx[expo] = PolynomialDegree(0)
    rule = PowerRule()
    result = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert result.is_polynomial() and result.degree == 8


def test_unary_function():
    rule = UnaryFunctionRule()
    ctx = {}
    result = rule.checked_apply(PE(ET.UnaryFunction), ctx)
    assert not result.is_polynomial()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
import pytest
import hypothesis.strategies as st
from suspect import set_pyomo4_expression_tree
from suspect.interval import Interval
from suspect.monotonicity.monotonicity import Monotonicity
from suspect.convexity.convexity import Convexity
from suspect.context import SpecialStructurePropagationContext


@st.composite
def reals(draw, min_value=None, max_value=None, allow_infinity=True):
    if min_value is not None and max_value is not None:
        allow_infinity = False
    return draw(st.floats(
        min_value=min_value, max_value=max_value,
        allow_nan=False, allow_infinity=allow_infinity
    ))


@st.composite
def coefficients(draw, min_value=None, max_value=None):
    return draw(st.floats(
        min_value=min_value, max_value=max_value,
        allow_nan=False, allow_infinity=False,
    ))


class PlaceholderExpression(object):
    depth = 0
    def __init__(self, expression_type=None, children=None, coefficients=None,
                 is_constant=False, value=None, constant_term=None,
                 bounded_above=False, bounded_below=False,
                 lower_bound=None, upper_bound=None,
                 is_minimizing=False, func_type=None):
        self.expression_type = expression_type
        self.children = children
        if coefficients is not None:
            self._coefficients = dict([(ch, co) for ch, co in zip(children, coefficients)])
        else:
            self._coefficients = {}
        self.is_constant = lambda: is_constant
        self.constant_term = constant_term
        self.value = value
        self.bounded_above = lambda: bounded_above
        self.bounded_below = lambda: bounded_below
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.is_minimizing = lambda: is_minimizing
        self.func_type = func_type

    def coefficient(self, expr):
        return self._coefficients[expr]

@pytest.fixture
def ctx():
    return SpecialStructurePropagationContext()


def bound_description_to_bound(bound_str):
    if isinstance(bound_str, str):
        return {
            'zero': Interval.zero(),
            'nonpositive': Interval(None, 0),
            'nonnegative': Interval(0, None),
            'positive': Interval(1, None),
            'negative': Interval(None, -1),
            'unbounded': Interval(None, None),
        }[bound_str]
    elif isinstance(bound_str, Interval):
        return bound_str
    else:
        return Interval(bound_str, bound_str)


def mono_description_to_mono(mono_str):
    return {
        'nondecreasing': Monotonicity.Nondecreasing,
        'nonincreasing': Monotonicity.Nonincreasing,
        'constant': Monotonicity.Constant,
        'unknown': Monotonicity.Unknown,
    }[mono_str]


def cvx_description_to_cvx(cvx_str):
    return {
        'convex': Convexity.Convex,
        'concave': Convexity.Concave,
        'linear': Convexity.Linear,
        'unknown': Convexity.Unknown,
    }[cvx_str]


def pytest_sessionstart(session):
    set_pyomo4_expression_tree()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import suspect.dag.expressions as dex
from tests.conftest import PlaceholderExpression


class TestConstraintLinearComponent(object):
    def test_is_only_child(self):
        root = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.linear_component() == root

    def test_no_linear_expression(self):
        root = dex.SumExpression(children=[PlaceholderExpression()])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.linear_component() == None

    def test_as_child_of_sum(self):
        linear = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        root = dex.SumExpression(children=[PlaceholderExpression(), linear, PlaceholderExpression()])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.linear_component() == linear

    def test_malformed_expression(self):
        linear_0 = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        linear_1 = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        root = dex.SumExpression(children=[PlaceholderExpression(), linear_0, linear_1])
        constr = dex.Constraint('c0', 0, 1, [root])
        with pytest.raises(AssertionError):
            constr.linear_component()


class TestConstraintNonlinearComponent(object):
    def test_is_only_child(self):
        root = dex.SinExpression([PlaceholderExpression()])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.nonlinear_component() == [root]

    def test_no_nonlinear_expression(self):
        root = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.nonlinear_component() == None

    def test_as_child_of_sum(self):
        linear = dex.LinearExpression(coefficients=[1.0], children=[PlaceholderExpression()])
        nonlinear_0 = dex.ExpExpression([PlaceholderExpression()])
        nonlinear_1 = dex.LogExpression([PlaceholderExpression()])
        root = dex.SumExpression(children=[nonlinear_0, linear, nonlinear_1])
        constr = dex.Constraint('c0', 0, 1, [root])
        assert constr.nonlinear_component() == [nonlinear_0, nonlinear_1]



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
import pyomo.environ as aml
from suspect.pyomo.expr_visitor import (
    InequalityExpression,
    EqualityExpression,
)
from suspect.interval import Interval

numeric_types = (int, Number, aml.NumericConstant)


def numeric_value(n):
    if isinstance(n, (int, Number)):
        return n
    elif isinstance(n, aml.NumericConstant):
        return n.value
    else:
        raise ValueError('must be one of numeric_types')


def bounds_and_expr(expr):
    if isinstance(expr, InequalityExpression):
        return _inequality_bounds_and_expr(expr)
    elif isinstance(expr, EqualityExpression):
        return _equality_bounds_and_expr(expr)
    else:
        raise ValueError('expr must be InequalityExpression or EqualityExpression')


def _inequality_bounds_and_expr(expr):
    if len(expr._args) == 2:
        (lhs, rhs) = expr._args
        if isinstance(lhs, aml.NumericConstant):
            return Interval(lhs.value, None), rhs
        else:
            return Interval(None, rhs.value), lhs
    elif len(expr._args) == 3:
        (lhs, ex, rhs) = expr._args
        return Interval(lhs.value, rhs.value), ex
    else:
        raise ValueError('Malformed InequalityExpression')


def _equality_bounds_and_expr(expr):
    if len(expr._args) == 2:
        body, rhs = expr._args
        return Interval(rhs.value, rhs.value), body
    else:
        raise ValueError('Malformed EqualityExpression')


def model_variables(model):
    """Return a list of variables in the model"""
    for variables in model.component_map(aml.Var, active=True).itervalues():
        for idx in variables:
            yield variables[idx]


def model_constraints(model):
    """Return a list of constraints in the model"""
    for cons in model.component_map(aml.Constraint, active=True).itervalues():
        for idx in cons:
            yield cons[idx]


def model_objectives(model):
    """Return a list of objectives in the model"""
    for obj in model.component_map(aml.Objective, active=True).itervalues():
        for idx in obj:
            yield obj[idx]



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number
import pyomo.environ as aml
from suspect.pyomo.util import (
    model_variables,
    model_objectives,
    model_constraints,
    bounds_and_expr,
)
from suspect.pyomo.expr_visitor import ExpressionHandler, bottom_up_visit
from suspect.pyomo.expr_dict import ExpressionDict
from suspect.float_hash import BTreeFloatHasher
from suspect.dag.dag import ProblemDag
import suspect.dag.expressions as dex


def dag_from_pyomo_model(model):
    """Convert the Pyomo ``model`` to SUSPECT DAG.

    Parameters
    ----------
    model : ConcreteModel
        the Pyomo model.

    Returns
    -------
    ProblemDag
        SUSPECT problem DAG.
    """
    dag = ProblemDag(name=model.name)
    factory = ComponentFactory(dag)
    for omo_var in model_variables(model):
        new_var = factory.variable(omo_var)
        dag.add_variable(new_var)

    for omo_cons in model_constraints(model):
        new_cons = factory.constraint(omo_cons)
        dag.add_constraint(new_cons)

    for omo_obj in model_objectives(model):
        new_obj = factory.objective(omo_obj)
        dag.add_objective(new_obj)

    return dag


def convert_domain(dom):
    if isinstance(dom, aml.RealSet):
        return dex.Domain.REALS
    elif isinstance(dom, aml.IntegerSet):
        return dex.Domain.INTEGERS
    elif isinstance(dom, aml.BooleanSet):
        return dex.Domain.BINARY


def convert_expression(memo, dag, expr):
    handler = ExpressionConverterHandler(memo, dag)
    bottom_up_visit(handler, expr)
    return memo[expr]


class ExpressionConverterHandler(ExpressionHandler):
    def __init__(self, memo, dag):
        self.memo = memo
        self.dag = dag

    def get(self, expr):
        if isinstance(expr, Number):
            const = aml.NumericConstant(expr)
            return self.get(const)
        else:
            return self.memo[expr]

    def set(self, expr, new_expr):
        self.memo[expr] = new_expr
        if len(new_expr.children) > 0:
            for child in new_expr.children:
                child.add_parent(new_expr)
        self.dag.add_vertex(new_expr)

    def _check_children(self, expr):
        for a in expr._args:
            if self.get(a) is None:
                raise RuntimeError('unknown child')

    def visit_number(self, n):
        const = aml.NumericConstant(n)
        self.visit_numeric_constant(const)

    def visit_numeric_constant(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]
        const = dex.Constant(expr.value)
        self.set(expr, const)

    def visit_variable(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]
        raise AssertionError('Unknown variable encountered')

    def visit_equality(self, expr):
        raise AssertionError('Invalid EqualityExpression encountered')

    def visit_inequality(self, expr):
        raise AssertionError('Invalid EqualityExpression encountered')

    def visit_product(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]
        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.ProductExpression(children)
        self.set(expr, new_expr)

    def visit_division(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.DivisionExpression(children)
        self.set(expr, new_expr)

    def visit_sum(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.SumExpression(children)
        self.set(expr, new_expr)

    def visit_linear(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        coeffs = [expr._coef[id(a)] for a in expr._args]
        const = expr._const
        new_expr = dex.LinearExpression(
            coeffs, children, const
        )
        self.set(expr, new_expr)

    def visit_negation(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.NegationExpression(children)
        self.set(expr, new_expr)

    def visit_unary_function(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        assert len(children) == 1
        fun = expr.name
        ExprClass = {
            'sqrt': dex.SqrtExpression,
            'exp': dex.ExpExpression,
            'log': dex.LogExpression,
            'sin': dex.SinExpression,
            'cos': dex.CosExpression,
            'tan': dex.TanExpression,
            'asin': dex.AsinExpression,
            'acos': dex.AcosExpression,
            'atan': dex.AtanExpression,
        }.get(fun)
        if ExprClass is None:
            raise AssertionError('Unknwon function', fun)
        new_expr = ExprClass(children)
        self.set(expr, new_expr)


    def visit_abs(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.AbsExpression(children)
        self.set(expr, new_expr)

    def visit_pow(self, expr):
        if self.memo[expr] is not None:
            return self.memo[expr]

        self._check_children(expr)
        children = [self.get(a) for a in expr._args]
        new_expr = dex.PowExpression(children)
        self.set(expr, new_expr)


class ComponentFactory(object):
    def __init__(self, dag):
        self._components = ExpressionDict(float_hasher=BTreeFloatHasher())
        self.dag = dag

    def variable(self, omo_var):
        comp = self._components.get(omo_var)
        if comp is not None:
            return comp
        domain = convert_domain(omo_var.domain)
        new_var = dex.Variable(omo_var.name, omo_var.lb, omo_var.ub, domain)
        self._components[omo_var] = new_var
        return new_var

    def constraint(self, omo_cons):
        bounds, expr = bounds_and_expr(omo_cons.expr)
        new_expr = self.expression(expr)
        constraint = dex.Constraint(
            omo_cons.name,
            bounds.lower_bound,
            bounds.upper_bound,
            [new_expr],
        )
        new_expr.add_parent(constraint)
        return constraint

    def objective(self, omo_obj):
        if omo_obj.is_minimizing():
            sense = dex.Sense.MINIMIZE
        else:
            sense = dex.Sense.MAXIMIZE

        new_expr = self.expression(omo_obj.expr)
        obj = dex.Objective(
            omo_obj.name, sense=sense, children=[new_expr]
        )
        new_expr.add_parent(obj)
        return obj

    def expression(self, expr):
        return convert_expression(self._components, self.dag, expr)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Iterators over vertices of ProblemDag."""
import abc
from suspect.interfaces import ForwardIterator, BackwardIterator
from suspect.dag.dag import ProblemDag
from suspect.dag.expressions import Expression
from suspect.dag.vertices_list import VerticesList


class _DagIterator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _get_next_vertices(self, expr):
        pass

    @abc.abstractproperty
    def _reverse(self) -> bool:
        pass

    def _iterate(self, _problem, visitor, ctx, starting_vertices):
        changed_vertices = []
        vertices = VerticesList(starting_vertices, reverse=self._reverse)
        seen = set()
        while vertices:
            curr_vertex = vertices.pop()
            if id(curr_vertex) in seen:
                continue
            has_changes = visitor.visit(curr_vertex, ctx)
            seen.add(id(curr_vertex))

            if has_changes:
                changed_vertices.append(curr_vertex)
                for vertex in self._get_next_vertices(curr_vertex):
                    if id(vertex) not in seen:
                        vertices.append(vertex)

        return changed_vertices


class DagForwardIterator(_DagIterator, ForwardIterator):
    """Forward iterator over suspect.dag.ProblemDag vertices."""
    def _get_next_vertices(self, expr):
        return expr.parents

    @property
    def _reverse(self) -> bool:
        return False

    def iterate(self, problem, visitor, ctx, *_args, **kwargs): # pylint: disable=missing-docstring
        starting_vertices = kwargs.pop('starting_vertices', None)
        # pylint: disable=protected-access
        if starting_vertices is None:
            starting_vertices = problem._sources
        else:
            starting_vertices = problem._sources + starting_vertices

        return super()._iterate(
            problem,
            visitor,
            ctx,
            starting_vertices=starting_vertices,
        )


class DagBackwardIterator(_DagIterator, BackwardIterator):
    """Backward iterator over suspect.dag.ProblemDag vertices."""
    def _get_next_vertices(self, expr):
        return expr.children

    @property
    def _reverse(self) -> bool:
        return True

    def iterate(self, problem, visitor, ctx, *_args, **kwargs): # pylint: disable=missing-docstring
        starting_vertices = kwargs.pop('starting_vertices', None)
        # pylint: disable=protected-access
        if starting_vertices is None:
            starting_vertices = problem._sinks
        else:
            starting_vertices = problem._sinks + starting_vertices

        return super()._iterate(
            problem,
            visitor,
            ctx,
            starting_vertices=starting_vertices,
        )



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visitor applying rules for monotonicity propagation."""
from suspect.interfaces import CombineUnaryFunctionRules
from suspect.visitor import Visitor
from suspect.monotonicity.rules import * # pylint: disable=wildcard-import


class MonotonicityPropagationVisitor(Visitor):
    """Visitor applying monotonicity rules."""
    def handle_result(self, expr, result, ctx):
        ctx.set_monotonicity(expr, result)
        return True

    def register_rules(self):
        return [
            VariableRule(),
            ConstantRule(),
            ConstraintRule(),
            ObjectiveRule(),
            DivisionRule(),
            ProductRule(),
            LinearRule(),
            QuadraticRule(),
            SumRule(),
            NegationRule(),
            PowerRule(),
            CombineUnaryFunctionRules(
                AbsRule(),
                SqrtRule(),
                ExpRule(),
                LogRule(),
                TanRule(),
                AtanRule(),
                SinRule(),
                AsinRule(),
                CosRule(),
                AcosRule(),
            )
        ]



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file
import pytest
from unittest.mock import MagicMock
from hypothesis import given, assume
import hypothesis.strategies as st
import suspect.dag.expressions as dex
from suspect.interval import Interval as I
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.fbbt.tightening.rules import *
from tests.conftest import (
    PlaceholderExpression as PE,
    bound_description_to_bound,
    mono_description_to_mono,
    coefficients,
    reals,
)


class TighteningContext:
    def __init__(self, bounds=None):
        if bounds is None:
            bounds = {}
        self._bounds = bounds

    def bounds(self, expr):
        return self._bounds[expr]


@given(reals(), reals())
def test_constraint_rule(bound1, bound2):
    lower_bound = min(bound1, bound2)
    upper_bound = max(bound1, bound2)
    rule = ConstraintRule()
    ctx = TighteningContext()
    child = PE()
    result = rule.checked_apply(
        PE(ET.Constraint, [child], lower_bound=lower_bound, upper_bound=upper_bound),
        ctx,
    )
    assert result[child] == I(lower_bound, upper_bound)


class TestSumRule:
    children = None
    expr = None

    def _rule_result(self, children_bounds, expr_bounds):
        children = [PE() for _ in children_bounds]
        expr = PE(ET.Sum, children)

        starting_bounds = {expr: expr_bounds}
        for child, bounds in zip(children, children_bounds):
            starting_bounds[child] = bounds
        ctx = TighteningContext(bounds=starting_bounds)
        rule = SumRule()
        self.children = children
        self.expr = expr
        return rule.checked_apply(expr, ctx)

    def test_unbounded_expr(self):
        new_bounds = self._rule_result([I(0, 1), I(-1, 0)], I(0, None))
        assert new_bounds is None

    def test_bounded_expr(self):
        expr_bounds = I(-5 , 5)
        children_bounds = [I(1, 2), I(-2, 2), I(-10, 10)]
        new_bounds = self._rule_result(children_bounds, expr_bounds)
        assert new_bounds[self.children[2]] == I(-9, 6)


class TestLinearRule:
    children = None
    expr = None

    def _rule_result(self, coefficients, children_bounds, expr_bounds):
        children = [PE() for _ in children_bounds]
        expr = PE(ET.Linear, children, coefficients=coefficients, constant_term=0)

        starting_bounds = {expr: expr_bounds}
        for child, bounds in zip(children, children_bounds):
            starting_bounds[child] = bounds
        ctx = TighteningContext(bounds=starting_bounds)
        rule = LinearRule()
        self.children = children
        self.expr = expr
        return rule.checked_apply(expr, ctx)

    def test_unbounded_expr(self):
        new_bounds = self._rule_result([-1, 1], [I(0, 1), I(-1, 0)], I(0, None))
        assert new_bounds is None

    def test_bounded_expr(self):
        expr_bounds = I(-5 , 5)
        children_bounds = [I(-2, -1), I(-2, 2), I(-10, 10)]
        new_bounds = self._rule_result([-1, 1, 2], children_bounds, expr_bounds)
        assert new_bounds[self.children[2]] == I(-4.5, 3)


class TestPowerRuleConstantExpo:
    base = None

    def _rule_result(self, expo, expr_bounds):
        base = PE()
        expo = PE(ET.Constant, value=expo, is_constant=True)
        expr = PE(ET.Power, [base, expo])
        rule = PowerRule()
        self.base = base
        ctx = TighteningContext(bounds={expr: expr_bounds})
        return rule.checked_apply(expr, ctx)

    def test_square(self):
        new_bounds = self._rule_result(2.0, I(0, 4))
        assert new_bounds[self.base] == I(-2, 2)

    def test_non_square(self):
        new_bounds = self._rule_result(3.0, I(0, 4))
        assert new_bounds is None


class TestUnboundedFunctionsRule:
    @pytest.mark.parametrize('expr_type,rule_cls', [
        (UFT.Sin, SinRule), (UFT.Cos, CosRule), (UFT.Tan, TanRule),
        (UFT.Asin, AsinRule), (UFT.Acos, AcosRule), (UFT.Atan, AtanRule)
    ])
    @given(a=reals(), b=reals())
    def test_unbounded(self, a, b, expr_type, rule_cls):
        lower = min(a, b)
        upper = max(a, b)
        rule = rule_cls()
        child = PE()
        expr = PE(ET.UnaryFunction, [child], func_type=expr_type)
        ctx = TighteningContext({expr: I(lower, upper)})
        new_bounds = rule.checked_apply(expr, ctx)
        assert new_bounds[child] == I(None, None)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds propagation rules."""
import operator
from functools import reduce
from suspect.interval import Interval
from suspect.interfaces import Rule
from suspect.expression import ExpressionType, UnaryFunctionType
from suspect.math import almosteq # pylint: disable=no-name-in-module


class VariableRule(Rule):
    """Bound propagation rule for variables."""
    root_expr = ExpressionType.Variable

    def apply(self, expr, _ctx):
        return Interval(expr.lower_bound, expr.upper_bound)


class ConstantRule(Rule):
    """Bound propagation rule for constants."""
    root_expr = ExpressionType.Constant

    def apply(self, expr, _ctx):
        return Interval(expr.value, expr.value)


class ConstraintRule(Rule):
    """Bound propagation rule for constraints."""
    root_expr = ExpressionType.Constraint

    def apply(self, expr, ctx):
        child = expr.children[0]
        child_bounds = ctx.bounds(child)
        constraint_bounds = Interval(expr.lower_bound, expr.upper_bound)
        return constraint_bounds.intersect(child_bounds)


class ObjectiveRule(Rule):
    """Bound propagation rule for objectives."""
    root_expr = ExpressionType.Objective

    def apply(self, expr, ctx):
        child = expr.children[0]
        return ctx.bounds(child)


class ProductRule(Rule):
    """Bound propagation rule for products."""
    root_expr = ExpressionType.Product

    def apply(self, expr, ctx):
        children_bounds = [ctx.bounds(child) for child in expr.children]
        return reduce(operator.mul, children_bounds, 1.0)


class QuadraticRule(Rule):
    """Bound propagation rule for quadratic."""
    root_expr = ExpressionType.Quadratic

    def apply(self, expr, ctx):
        return sum([
            ctx.bounds(term.var1) * ctx.bounds(term.var2) * term.coefficient
            for term in expr.terms
        ])


class DivisionRule(Rule):
    """Bound propagation rule for divisions."""
    root_expr = ExpressionType.Division

    def apply(self, expr, ctx):
        num, den = expr.children
        return ctx.bounds(num) / ctx.bounds(den)


class LinearRule(Rule):
    """Bound propagation rule for linear expressions."""
    root_expr = ExpressionType.Linear

    def apply(self, expr, ctx):
        children_contribution = sum(
            expr.coefficient(child) * ctx.bounds(child)
            for child in expr.children
        )
        constant_contribution = Interval(expr.constant_term, expr.constant_term)
        return children_contribution + constant_contribution


class SumRule(Rule):
    """Bound propagation rule for sum."""
    root_expr = ExpressionType.Sum

    def apply(self, expr, ctx):
        return sum(ctx.bounds(child) for child in expr.children)


class PowerRule(Rule):
    """Bound propagation rule for power."""
    root_expr = ExpressionType.Power

    def apply(self, expr, ctx):
        _, expo = expr.children
        if not expo.is_constant():
            return Interval(None, None)

        is_even = almosteq(expo.value % 2, 0)
        is_positive = expo.value > 0
        if is_even and is_positive:
            return Interval(0, None)
        return Interval(None, None)


class NegationRule(Rule):
    """Bound propagation rule for negation."""
    root_expr = ExpressionType.Negation

    def apply(self, expr, ctx):
        return -ctx.bounds(expr.children[0])


class UnaryFunctionRule(Rule):
    """Bound propagation rule for unary functions."""
    root_expr = ExpressionType.UnaryFunction

    _FUNC_TYPE_TO_NAME = {
        UnaryFunctionType.Abs: 'abs',
        UnaryFunctionType.Sqrt: 'sqrt',
        UnaryFunctionType.Exp: 'exp',
        UnaryFunctionType.Log: 'log',
        UnaryFunctionType.Sin: 'sin',
        UnaryFunctionType.Cos: 'cos',
        UnaryFunctionType.Tan: 'tan',
        UnaryFunctionType.Asin: 'asin',
        UnaryFunctionType.Acos: 'acos',
        UnaryFunctionType.Atan: 'atan',
    }

    def apply(self, expr, ctx):
        child = expr.children[0]
        child_bounds = ctx.bounds(child)
        func_name = self._func_name(expr.func_type)
        func = getattr(child_bounds, func_name)
        return func()

    def _func_name(self, func_type):
        func_name = self._FUNC_TYPE_TO_NAME.get(func_type)
        if func_name is None:
            raise RuntimeError('Invalid function type {}'.format(func_type))
        return func_name



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds tightening rules."""
import numpy as np
from suspect.interfaces import Rule, UnaryFunctionRule
from suspect.expression import ExpressionType, UnaryFunctionType
from suspect.interval import Interval
from suspect.math import almosteq, inf # pylint: disable=no-name-in-module


MAX_EXPR_CHILDREN = 1000


class ConstraintRule(Rule):
    """Return new bounds for constraint."""
    root_expr = ExpressionType.Constraint

    def apply(self, expr, ctx):
        child = expr.children[0]
        bounds = Interval(expr.lower_bound, expr.upper_bound)
        return {
            child: bounds
        }


class SumRule(Rule):
    """Return new bounds for sum."""
    root_expr = ExpressionType.Sum

    def apply(self, expr, ctx):
        expr_bound = ctx.bounds(expr)
        if expr_bound.size() == inf:
            return None
        if len(expr.children) > MAX_EXPR_CHILDREN: # pragma: no cover
            return None
        child_bounds = {}
        for child, siblings in _sum_child_and_siblings(expr.children):
            siblings_bound = sum(ctx.bounds(s) for s in siblings)
            child_bounds[child] = expr_bound - siblings_bound
        return child_bounds


class LinearRule(Rule):
    """Return new bounds for linear expressions."""
    root_expr = ExpressionType.Linear

    def apply(self, expr, ctx):
        expr_bound = ctx.bounds(expr)
        if expr_bound.size() == inf:
            return None
        if len(expr.children) > MAX_EXPR_CHILDREN: # pragma: no cover
            return None
        child_bounds = {}
        const = expr.constant_term
        coefficients = [expr.coefficient(ch) for ch in expr.children]
        for (child_c, child), siblings in _linear_child_and_siblings(coefficients, expr.children):
            siblings_bound = sum(ctx.bounds(s) * c for c, s in siblings) + const
            child_bounds[child] = (expr_bound - siblings_bound) / child_c
        return child_bounds


class PowerRule(Rule):
    """Return new bounds for power expressions."""
    root_expr = ExpressionType.Power

    def apply(self, expr, ctx):
        base, expo = expr.children
        if not expo.is_constant():
            return None
        if not almosteq(expo.value, 2):
            return None

        bounds = ctx.bounds(expr)
        # the bound of a square number is never negative, but check anyway to
        # avoid unexpected crashes.
        if not bounds.is_nonnegative():
            return None

        sqrt_bound = bounds.sqrt()
        return {
            base: Interval(-sqrt_bound.upper_bound, sqrt_bound.upper_bound)
        }


class _UnaryFunctionRule(UnaryFunctionRule):
    def apply(self, expr, ctx):
        child = expr.children[0]
        bounds = ctx.bounds(expr)
        return {
            child: self._child_bounds(bounds)
        }

    def _child_bounds(self, bounds):
        pass


class _BoundedFunctionRule(_UnaryFunctionRule):
    func_name = None

    def _child_bounds(self, bounds):
        func = getattr(bounds, self.func_name)
        return func.inverse()


class AbsRule(_BoundedFunctionRule):
    """Return new bounds for abs."""
    func_name = 'abs'


class SqrtRule(_BoundedFunctionRule):
    """Return new bounds for sqrt."""
    func_name = 'sqrt'


class ExpRule(_BoundedFunctionRule):
    """Return new bounds for exp."""
    func_name = 'exp'


class LogRule(_BoundedFunctionRule):
    """Return new bounds for log."""
    func_name = 'log'


class _UnboundedFunctionRule(_UnaryFunctionRule):
    def _child_bounds(self, bounds):
        return Interval(None, None)


class SinRule(_UnboundedFunctionRule):
    """Return new bounds for sin."""
    func_type = UnaryFunctionType.Sin


class CosRule(_UnboundedFunctionRule):
    """Return new bounds for cos."""
    func_type = UnaryFunctionType.Cos


class TanRule(_UnboundedFunctionRule):
    """Return new bounds for tan."""
    func_type = UnaryFunctionType.Tan


class AsinRule(_UnboundedFunctionRule):
    """Return new bounds for asin."""
    func_type = UnaryFunctionType.Asin


class AcosRule(_UnboundedFunctionRule):
    """Return new bounds for acos."""
    func_type = UnaryFunctionType.Acos


class AtanRule(_UnboundedFunctionRule):
    """Return new bounds for atan."""
    func_type = UnaryFunctionType.Atan


def _sum_child_and_siblings(children):
    for i, _ in enumerate(children):
        yield children[i], children[:i] + children[i+1:]


def _linear_child_and_siblings(coefficients, children):
    for i, child in enumerate(children):
        child_c = coefficients[i]
        other_children = np.concatenate((children[:i], children[i+1:]))
        other_coefficients = np.concatenate((coefficients[:i], coefficients[i+1:]))
        yield (child_c, child), zip(other_coefficients, other_children)



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import pyomo.environ as aml
import numpy as np
from suspect.dag import ProblemDag
import suspect.dag.expressions as dex
from suspect.pyomo.convert import ComponentFactory, dag_from_pyomo_model
from suspect.pyomo.util import model_variables, model_constraints
from suspect.math.arbitrary_precision import inf


class TestConvertVariable(object):
    def test_continuous_variables(self):
        m = aml.ConcreteModel()
        # 10 continuous variables in [-inf, inf]
        m.x = aml.Var(range(10))

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('x')
            assert new_var.lower_bound is None
            assert new_var.upper_bound is None
            assert new_var.domain == dex.Domain.REALS
            count += 1
        assert count == 10

    def test_integer_variables(self):
        m = aml.ConcreteModel()
        # 5 integer variables in [-10, 5]
        m.y = aml.Var(range(5), bounds=(-10, 5), domain=aml.Integers)

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('y')
            assert new_var.lower_bound == -10
            assert new_var.upper_bound == 5
            assert new_var.domain == dex.Domain.INTEGERS
            count += 1
        assert count == 5

    def test_binary_variables(self):
        m = aml.ConcreteModel()
        # 10 binary variables
        m.b = aml.Var(range(10), domain=aml.Binary)

        dag = ProblemDag()
        factory = ComponentFactory(dag)
        count = 0
        for omo_var in model_variables(m):
            new_var = factory.variable(omo_var)
            assert new_var.name.startswith('b')
            assert new_var.lower_bound == 0
            assert new_var.upper_bound == 1
            assert new_var.domain == dex.Domain.BINARY
            count += 1
        assert count == 10


class TestConvertExpression(object):
    def test_simple_model(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c = aml.Constraint(m.I, rule=lambda m, i: m.x[i] + 2 >= 0)

        dag = dag_from_pyomo_model(m)

        assert len(dag.constraints) == 10
        for constraint in dag.constraints.values():
            assert constraint.lower_bound == 0.0
            assert constraint.upper_bound == inf
            assert len(constraint.children) == 1
            root = constraint.children[0]
            assert isinstance(root, dex.LinearExpression)
            assert len(root.children) == 1
            assert root.constant_term == 2.0
            assert isinstance(root.children[0], dex.Variable)

    def test_nested_expressions(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)
        m.y = aml.Var(m.I)

        m.c = aml.Constraint(m.I, rule=lambda m, i: aml.sin(2*m.x[i] - m.y[i]) / (m.x[i] + 1) <= 100)

        dag = dag_from_pyomo_model(m)

        assert len(dag.constraints) == 10
        for constraint in dag.constraints.values():
            assert constraint.lower_bound == -inf
            assert constraint.upper_bound == 100
            assert len(constraint.children) == 1
            root = constraint.children[0]
            assert isinstance(root, dex.DivisionExpression)
            num, den = root.children
            assert isinstance(num, dex.SinExpression)
            assert len(num.children) == 1
            num_inner = num.children[0]
            assert isinstance(num_inner, dex.LinearExpression)
            assert np.isclose(2.0, num_inner.coefficient(num_inner.children[0]))
            assert np.isclose(-1.0, num_inner.coefficient(num_inner.children[1]))
            assert isinstance(den, dex.LinearExpression)
            assert den.constant_term == 1.0

    def test_product(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c = aml.Constraint(range(9), rule=lambda m, i: 2*m.x[i] * 3*m.x[i+1] >= 0)

        dag = dag_from_pyomo_model(m)

        assert len(dag.constraints) == 9
        for constraint in dag.constraints.values():
            root = constraint.children[0]
            assert isinstance(root, dex.ProductExpression)
            assert len(root.children) == 2
            linear = root.children[0]
            var = root.children[1]
            assert isinstance(var, dex.Variable)
            assert isinstance(linear, dex.LinearExpression)
            assert np.isclose(6.0, linear.coefficient(linear.children[0]))

    def test_sum(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c = aml.Constraint(range(8), rule=lambda m, j: sum(m.x[i]*m.x[i+1] for i in range(j+2)) >= 0)

        dag = dag_from_pyomo_model(m)

        assert len(dag.constraints) == 8
        for constraint in dag.constraints.values():
            root = constraint.children[0]
            assert isinstance(root, dex.SumExpression)
            for c in root.children:
                assert isinstance(c, dex.ProductExpression)

    def test_negation(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c = aml.Constraint(expr=-aml.cos(m.x[0]) >= 0)

        dag = dag_from_pyomo_model(m)

        constraint = dag.constraints['c']
        root = constraint.children[0]
        assert isinstance(root, dex.NegationExpression)

    def test_abs(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c = aml.Constraint(expr=abs(m.x[0]) >= 0)

        dag = dag_from_pyomo_model(m)

        constraint = dag.constraints['c']
        root = constraint.children[0]
        assert isinstance(root, dex.AbsExpression)

    def test_pow(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)

        m.c0 = aml.Constraint(expr=aml.cos(m.x[0])**2.0 >= 1)
        m.c1 = aml.Constraint(expr=2**aml.sin(m.x[1]) >= 1)

        dag = dag_from_pyomo_model(m)

        c0 = dag.constraints['c0']
        root_c0 = c0.children[0]
        assert isinstance(root_c0, dex.PowExpression)
        assert len(root_c0.children) == 2
        assert isinstance(root_c0.children[0], dex.CosExpression)
        assert isinstance(root_c0.children[1], dex.Constant)
        assert root_c0.children[1].value == 2.0

        c1 = dag.constraints['c1']
        root_c1 = c1.children[0]
        assert isinstance(root_c1, dex.PowExpression)
        assert len(root_c1.children) == 2
        assert isinstance(root_c1.children[0], dex.Constant)
        assert isinstance(root_c1.children[1], dex.SinExpression)


class TestConvertObjective(object):
    def test_min(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)
        m.obj = aml.Objective(expr=sum(m.x[i] for i in m.I))

        dag = dag_from_pyomo_model(m)
        assert len(dag.objectives) == 1
        obj = dag.objectives['obj']
        assert isinstance(obj.children[0], dex.LinearExpression)
        assert obj.sense == dex.Sense.MINIMIZE

    def test_max(self):
        m = aml.ConcreteModel()
        m.I = range(10)
        m.x = aml.Var(m.I)
        m.obj = aml.Objective(expr=sum(m.x[i] for i in m.I), sense=aml.maximize)

        dag = dag_from_pyomo_model(m)
        assert len(dag.objectives) == 1
        obj = dag.objectives['obj']
        assert isinstance(obj.children[0], dex.LinearExpression)
        assert obj.sense == dex.Sense.MAXIMIZE



# pylint: skip-file
import pytest
from hypothesis import assume, given
import hypothesis.strategies as st
import itertools
from tests.conftest import coefficients, reals
from tests.conftest import PlaceholderExpression as PE
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.monotonicity.monotonicity import Monotonicity as M
from suspect.interval import Interval as I
from suspect.monotonicity.rules import *
from suspect.math import almosteq, pi


class MonotonicityContext:
    def __init__(self, ctx=None, bounds=None):
        if ctx is None:
            ctx = {}
        if bounds is None:
            bounds = {}
        self.ctx = ctx
        self._bounds = bounds

    def monotonicity(self, expr):
        return self.ctx[expr]

    def bounds(self, expr):
        return self._bounds[expr]


@st.composite
def nondecreasing_terms(draw):
    coef = draw(coefficients())
    if coef > 0:
        return coef, M.Nondecreasing
    return coef, M.Nonincreasing


@st.composite
def nonincreasing_terms(draw):
    coef = draw(coefficients())
    if coef < 0:
        return coef, M.Nondecreasing
    return coef, M.Nonincreasing


@st.composite
def constant_terms(draw):
    coef = draw(coefficients())
    return coef, M.Constant


@st.composite
def nonnegative_cos_bounds(draw):
    start = draw(st.floats(
        min_value=1.6*pi,
        max_value=1.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-100, max_value=100)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonpositive_cos_bounds(draw):
    start = draw(st.floats(
        min_value=0.6*pi,
        max_value=1.5*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.5*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-10, max_value=10)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonnegative_sin_bounds(draw):
    start = draw(st.floats(
        min_value=0.1*pi,
        max_value=0.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=0.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-100, max_value=100)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonpositive_sin_bounds(draw):
    start = draw(st.floats(
        min_value=1.1*pi,
        max_value=1.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-10, max_value=10)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


def test_variable_is_nondecreasing():
    rule = VariableRule()
    ctx = MonotonicityContext()
    result = rule.checked_apply(PE(ET.Variable), ctx)
    assert result.is_nondecreasing() and (not result.is_constant())


def test_constant_is_constant():
    rule = ConstantRule()
    ctx = MonotonicityContext()
    result = rule.checked_apply(PE(ET.Constant), ctx)
    assert result.is_constant()


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
def test_lte_constraint(expr_mono, expected_mono):
    rule = ConstraintRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: expr_mono})
    result = rule.checked_apply(PE(ET.Constraint, [child], bounded_above=True), ctx)
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
def test_gte_constraint(expr_mono, expected_mono):
    rule = ConstraintRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: expr_mono})
    result = rule.checked_apply(PE(ET.Constraint, [child], bounded_below=True), ctx)
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Unknown),
    (M.Nonincreasing, M.Unknown),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
def test_eq_constraint(expr_mono, expected_mono):
    rule = ConstraintRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: expr_mono})
    result = rule.checked_apply(PE(ET.Constraint, [child], bounded_above=True, bounded_below=True), ctx)
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
def test_min_constraint(expr_mono, expected_mono):
    rule = ObjectiveRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: expr_mono})
    result = rule.checked_apply(PE(ET.Objective, [child], is_minimizing=True), ctx)
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
def test_max_constraint(expr_mono, expected_mono):
    rule = ObjectiveRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: expr_mono})
    result = rule.checked_apply(PE(ET.Objective, [child], is_minimizing=False), ctx)
    assert result == expected_mono


@pytest.mark.parametrize('mono_f,mono_g,bound_f,bound_g,expected', [
    (M.Nondecreasing, M.Constant, I(None, None), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(None, 0), M.Nondecreasing),

    (M.Nondecreasing, M.Constant, I(None, None), I(None, 0), M.Nonincreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(0, None), M.Nonincreasing),

    (M.Constant, M.Nondecreasing, I(0, None), I(None, None), M.Nondecreasing),
    (M.Constant, M.Nonincreasing, I(None, 0), I(None, None), M.Nondecreasing),

    (M.Constant, M.Nondecreasing, I(None, 0), I(None, None), M.Nonincreasing),
    (M.Constant, M.Nonincreasing, I(0, None), I(None, None), M.Nonincreasing),

    (M.Constant, M.Constant, I(None, None), I(None, None), M.Constant),

    (M.Nondecreasing, M.Nondecreasing, I(0, None), I(0, None), M.Nondecreasing),
    (M.Nondecreasing, M.Nonincreasing, I(None, 0), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Nondecreasing, I(0, None), I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing, I(None, 0), I(None, 0), M.Nondecreasing),

    (M.Nonincreasing, M.Nonincreasing, I(0, None), I(0, None), M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing, I(None, 0), I(0, None), M.Nonincreasing),
    (M.Nondecreasing, M.Nonincreasing, I(0, None), I(None, 0), M.Nonincreasing),
    (M.Nondecreasing, M.Nondecreasing, I(None, 0), I(None, 0), M.Nonincreasing),
])
def test_product(mono_f, mono_g, bound_f, bound_g, expected):
    rule = ProductRule()
    f = PE(ET.UnaryFunction)
    g = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({f: mono_f, g: mono_g}, {f: bound_f, g: bound_g})
    result = rule.checked_apply(PE(ET.Product, [f, g]), ctx)
    assert result == expected


@pytest.mark.parametrize('mono_f,mono_g,bound_f,bound_g,expected', [
    (M.Constant, M.Constant, I(1.0, 1.0), I(2.0, 2.0), M.Constant),
    (M.Constant, M.Constant, I(1.0, 1.0), I(0.0, 0.0), M.Unknown),

    (M.Nondecreasing, M.Constant, I(None, None), I(10.0, 10.0), M.Nondecreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(-1.0, -1.0), M.Nondecreasing),

    (M.Nondecreasing, M.Constant, I(None, None), I(-1.0, -1.0), M.Nonincreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(10.0, 10.0), M.Nonincreasing),

    (M.Nondecreasing, M.Nonincreasing, I(0, None), I(0, None), M.Nondecreasing),
    (M.Nondecreasing, M.Nondecreasing, I(None, 0), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing, I(0, None), I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, M.Nondecreasing, I(None, 0), I(None, 0), M.Nondecreasing),

    (M.Nonincreasing, M.Nondecreasing, I(0, None), I(0, None), M.Nonincreasing),
    (M.Nonincreasing, M.Nonincreasing, I(None, 0), I(0, None), M.Nonincreasing),
    (M.Nondecreasing, M.Nondecreasing, I(0, None), I(None, 0), M.Nonincreasing),
    (M.Nondecreasing, M.Nonincreasing, I(None, 0), I(None, 0), M.Nonincreasing),
])
def test_division(mono_f, mono_g, bound_f, bound_g, expected):
    rule = DivisionRule()
    f = PE(ET.UnaryFunction)
    g = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({f: mono_f, g: mono_g}, {f: bound_f, g: bound_g})
    result = rule.checked_apply(PE(ET.Division, [f, g]), ctx)
    assert result == expected



class TestLinear(object):
    def _result_with_terms(self, terms):
        rule = LinearRule()
        monotonicity = {}
        coefficients = [c for c, _ in terms]
        children = [PE(ET.Variable) for _ in terms]
        monos = [m for _, m in terms]
        for child, mono in zip(children, monos):
            monotonicity[child] = mono
        ctx = MonotonicityContext(monotonicity, {})
        result = rule.checked_apply(
            PE(ET.Linear, children=children, coefficients=coefficients),
            ctx,
        )
        return result

    @given(
        st.lists(
            st.one_of(nondecreasing_terms(), constant_terms()),
            min_size=1)
    )
    def test_nondecreasing(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_nondecreasing()

    @given(
        st.lists(
            st.one_of(nonincreasing_terms(), constant_terms()),
            min_size=1)
    )
    def test_nonincreasing(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_nonincreasing()

    @given(st.lists(constant_terms(), min_size=1))
    def test_constant(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_constant()

    @given(
        st.lists(nondecreasing_terms(), min_size=1),
        st.lists(nonincreasing_terms(), min_size=1),
    )
    def test_unknown(self, a, b):
        terms = a + b
        for c, _ in terms:
            assume(c != 0.0)
        mono = self._result_with_terms(terms)
        assert mono.is_unknown()



class TestSum(object):
    def _result_with_terms(self, monos):
        rule = SumRule()
        monotonicity = {}
        children = [PE(ET.UnaryFunction) for _ in monos]
        for child, mono in zip(children, monos):
            monotonicity[child] = mono
        ctx = MonotonicityContext(monotonicity, {})
        result = rule.checked_apply(PE(ET.Sum, children=children), ctx)
        return result

    @given(
        st.lists(st.just(M.Nondecreasing), min_size=1),
        st.lists(st.one_of(st.just(M.Nondecreasing), st.just(M.Constant))),
    )
    def test_nondecreasing(self, a, b):
        mono = self._result_with_terms(a + b)
        assert mono.is_nondecreasing()

    @given(
        st.lists(st.just(M.Nonincreasing), min_size=1),
        st.lists(st.one_of(st.just(M.Nonincreasing), st.just(M.Constant))),
    )
    def test_nonincreasing(self, a, b):
        mono = self._result_with_terms(a + b)
        assert mono.is_nonincreasing()

    @given(st.lists(st.just(M.Constant), min_size=1))
    def test_constant(self, a):
        mono = self._result_with_terms(a)
        assert mono.is_constant()

    @given(
        st.lists(st.just(M.Nondecreasing), min_size=1),
        st.lists(st.just(M.Nonincreasing), min_size=1),
    )
    def test_unknown(self, a, b):
        mono = self._result_with_terms(a + b)
        assert mono.is_unknown()


@pytest.mark.parametrize('mono,bounds,expected', [
    (M.Nondecreasing, I(0, None), M.Nondecreasing),
    (M.Nonincreasing, I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, I(0, None), M.Nonincreasing),
    (M.Nondecreasing, I(None, 0), M.Nonincreasing),
    (M.Constant, I(None, None), M.Constant),
    (M.Nondecreasing, I(None, None), M.Unknown),
    (M.Nonincreasing, I(None, None), M.Unknown),
])
def test_abs(mono, bounds, expected):
    rule = AbsRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: mono}, {child: bounds})
    result = rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=UFT.Abs), ctx)
    assert result == expected


@pytest.mark.parametrize('rule_cls,func_type', [
    (SqrtRule, UFT.Sqrt), (ExpRule, UFT.Exp), (LogRule, UFT.Log),
    (TanRule, UFT.Tan), (AsinRule, UFT.Asin), (AtanRule, UFT.Atan)
])
@pytest.mark.parametrize('mono,bounds', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
def test_nondecreasing_function(rule_cls, func_type, mono, bounds):
    rule = rule_cls()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: mono}, {child: bounds})
    result = rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=func_type), ctx)
    assert result == mono


@pytest.mark.parametrize('rule_cls,func_type', [(AcosRule, UFT.Acos)])
@pytest.mark.parametrize('mono,bounds', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
def test_nonincreasing_function(rule_cls, func_type, mono, bounds):
    rule = rule_cls()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: mono}, {child: bounds})
    result = rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=func_type), ctx)
    assert result == mono.negate()


@pytest.mark.parametrize('mono,bounds', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
def test_negation(mono, bounds):
    rule = NegationRule()
    child = PE(ET.UnaryFunction)
    ctx = MonotonicityContext({child: mono}, {child: bounds})
    result = rule.checked_apply(PE(ET.Negation, [child]), ctx)
    assert result == mono.negate()


class TestPowConstantBase(object):
    def _result_with_base_expo(self, base, mono_expo, bounds_expo):
        base = PE(ET.Constant, is_constant=True, value=base)
        expo = PE()
        rule = PowerRule()
        ctx = MonotonicityContext({
            base: M.Constant,
            expo: mono_expo,
        }, {
            base: I(base.value, base.value),
            expo: bounds_expo,
        })
        return rule.checked_apply(PE(ET.Power, [base, expo]), ctx)

    @pytest.mark.parametrize(
        'mono_expo,bounds_expo',
        itertools.product(
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)]
        )
    )
    @given(base=reals(max_value=-0.01, allow_infinity=False))
    def test_negative_base(self, base, mono_expo, bounds_expo):
        mono = self._result_with_base_expo(base, mono_expo, bounds_expo)
        assert mono == M.Unknown

    @pytest.mark.parametrize('mono_expo,bounds_expo,expected', [
        (M.Nondecreasing, I(None, 0), M.Nondecreasing),
        (M.Nondecreasing, I(0, None), M.Unknown),
        (M.Nonincreasing, I(0, None), M.Nondecreasing),
        (M.Nonincreasing, I(None, 0), M.Unknown),
    ])
    @given(base=reals(min_value=0.01, max_value=0.999))
    def test_base_between_0_and_1(self, base, mono_expo, bounds_expo, expected):
        mono = self._result_with_base_expo(base, mono_expo, bounds_expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_expo,bounds_expo,expected', [
        (M.Nondecreasing, I(0, None), M.Nondecreasing),
        (M.Nondecreasing, I(None, 0), M.Unknown),
        (M.Nonincreasing, I(None, 0), M.Nondecreasing),
        (M.Nonincreasing, I(0, None), M.Unknown),
    ])
    @given(base=reals(min_value=1.01, allow_infinity=False))
    def test_base_gt_1(self, base, mono_expo, bounds_expo, expected):
        mono = self._result_with_base_expo(base, mono_expo, bounds_expo)
        assert mono == expected


class TestPowConstantExponent(object):
    def _result_with_base_expo(self, mono_base, bounds_base, expo):
        base = PE()
        expo = PE(ET.Constant, is_constant=True, value=expo)
        rule = PowerRule()
        ctx = MonotonicityContext({
            base: mono_base,
            expo: M.Constant,
        }, {
            base: bounds_base,
            expo: I(expo.value, expo.value),
        })
        return rule.checked_apply(PE(ET.Power, [base, expo]), ctx)

    @pytest.mark.parametrize(
        'mono_base,bounds_base',
        itertools.product([M.Nonincreasing, M.Nondecreasing],
                          [I(None, 0), I(0, None), I(None, None)])
    )
    def test_exponent_equals_1(self, mono_base, bounds_base):
        mono = self._result_with_base_expo(mono_base, bounds_base, 1.0)
        assert mono == mono_base

    @pytest.mark.parametrize(
        'mono_base,bounds_base',
        itertools.product([M.Nonincreasing, M.Nondecreasing],
                          [I(None, 0), I(0, None), I(None, None)])
    )
    def test_exponent_equals_0(self, mono_base, bounds_base):
        mono = self._result_with_base_expo(mono_base, bounds_base, 0.0)
        assert mono == M.Constant

    @pytest.mark.parametrize('mono_base,bounds_base,expected', [
        (M.Nondecreasing, I(0, None), M.Nondecreasing),
        (M.Nonincreasing, I(None, 0), M.Nondecreasing),

        (M.Nondecreasing, I(None, 0), M.Nonincreasing),
        (M.Nonincreasing, I(0, None), M.Nonincreasing),
    ])
    @given(expo=st.integers(min_value=1))
    def test_positive_even_integer(self, expo, mono_base, bounds_base, expected):
        mono = self._result_with_base_expo(mono_base, bounds_base, 2*expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,bounds_base,expected', [
        (M.Nondecreasing, I(0, None), M.Nonincreasing),
        (M.Nonincreasing, I(None, 0), M.Nonincreasing),

        (M.Nondecreasing, I(None, 0), M.Nondecreasing),
        (M.Nonincreasing, I(0, None), M.Nondecreasing),
    ])
    @given(expo=st.integers(min_value=1))
    def test_negative_even_integer(self, expo, mono_base, bounds_base, expected):
        mono = self._result_with_base_expo(mono_base, bounds_base, -2*expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nondecreasing),
        (M.Nonincreasing, M.Nonincreasing),
    ])
    @given(expo=st.integers(min_value=1))
    def test_positive_odd_integer(self, expo, mono_base, expected):
        mono = self._result_with_base_expo(mono_base, I(None, None), 2*expo+1)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nonincreasing),
        (M.Nonincreasing, M.Nondecreasing),
    ])
    @given(expo=st.integers(min_value=1))
    def test_negative_odd_integer(self, expo, mono_base, expected):
        mono = self._result_with_base_expo(mono_base, I(None, None), -2*expo+1)
        assert mono == expected

    @pytest.mark.skip('Not updated to work with Quadratic')
    @pytest.mark.parametrize('mono_base', [M.Nondecreasing, M.Nondecreasing])
    @given(expo=reals(allow_infinity=False))
    def test_noninteger_negative_base(self, expo, mono_base):
        assume(not almosteq(expo,  0))
        assume(expo != int(expo))
        mono = self._result_with_base_expo(mono_base, I(None, 0), expo)
        assert mono == M.Unknown

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nondecreasing),
        (M.Nonincreasing, M.Nonincreasing)
    ])
    @given(expo=reals(allow_infinity=False, min_value=1e-5))
    def test_positive_noninteger(self, expo, mono_base, expected):
        mono = self._result_with_base_expo(mono_base, I(0, None), expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nonincreasing, M.Nondecreasing),
        (M.Nondecreasing, M.Nonincreasing)
    ])
    @given(expo=reals(allow_infinity=False, max_value=-1e-5))
    def test_negative_noninteger(self, expo, mono_base, expected):
        mono = self._result_with_base_expo(mono_base, I(0, None), expo)
        assert mono == expected

@pytest.mark.parametrize('mono_base,bounds_base,mono_expo,bounds_expo',
    itertools.product(
        [M.Nonincreasing, M.Nondecreasing, M.Unknown],
        [I(None, 0), I(0, None), I(None, None)],
        [M.Nonincreasing, M.Nondecreasing, M.Unknown],
        [I(None, 0), I(0, None), I(None, None)],
    )
)
def test_pow(mono_base, bounds_base, mono_expo, bounds_expo):
    base = PE()
    expo = PE()
    rule = PowerRule()
    ctx = MonotonicityContext({
        base: mono_base,
        expo: mono_expo,
    }, {
        base: bounds_base,
        expo: bounds_expo,
    })
    mono = rule.checked_apply(PE(ET.Power, [base, expo]), ctx)
    assert mono == M.Unknown


class TestSin(object):
    def _result_with_mono_bounds(self, mono, bounds):
        rule = SinRule()
        child = PE()
        ctx = MonotonicityContext({child: mono}, {child: bounds})
        return rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=UFT.Sin), ctx)

    @given(nonnegative_cos_bounds())
    def test_nondecreasing_nonnegative_cos(self, bounds):
        mono = self._result_with_mono_bounds(M.Nondecreasing, bounds)
        assert mono.is_nondecreasing()

    @given(nonnegative_cos_bounds())
    def test_nonincreasing_nonnegative_cos(self, bounds):
        mono = self._result_with_mono_bounds(M.Nonincreasing, bounds)
        assert mono.is_nonincreasing()

    @given(nonpositive_cos_bounds())
    def test_decreasing_nonpositive_cos(self, bounds):
        mono = self._result_with_mono_bounds(M.Nondecreasing, bounds)
        assert mono.is_nonincreasing()

    @given(nonpositive_cos_bounds())
    def test_nonincreasing_nonpositive_cos(self, bounds):
        mono = self._result_with_mono_bounds(M.Nonincreasing, bounds)
        assert mono.is_nondecreasing()


class TestCos(object):
    def _result_with_mono_bounds(self, mono, bounds):
        rule = CosRule()
        child = PE()
        ctx = MonotonicityContext({child: mono}, {child: bounds})
        return rule.checked_apply(PE(ET.UnaryFunction, [child], func_type=UFT.Cos), ctx)

    @given(nonnegative_sin_bounds())
    def test_nonincreasing_nonnegative_sin(self, bounds):
        mono = self._result_with_mono_bounds(M.Nonincreasing, bounds)
        assert mono.is_nondecreasing()

    @given(nonnegative_sin_bounds())
    def test_nondecreasing_nonnegative_sin(self, bounds):
        mono = self._result_with_mono_bounds(M.Nondecreasing, bounds)
        assert mono.is_nonincreasing()

    @given(nonpositive_sin_bounds())
    def test_nondecreasing_nonpositive_sin(self, bounds):
        mono = self._result_with_mono_bounds(M.Nondecreasing, bounds)
        assert mono.is_nondecreasing()

    @given(nonpositive_sin_bounds())
    def test_nonincreasing_nonpositive_sin(self, bounds):
        mono = self._result_with_mono_bounds(M.Nonincreasing, bounds)
        assert mono.is_nonincreasing()



#!/usr/bin/env python

# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import signal
import sys
import time
import logging
import json
import tempfile
import traceback
from mpmath import mpf
import boto3
from suspect.pyomo.osil_reader import read_osil
from suspect.pyomo.convert import dag_from_pyomo_model
from suspect import (
    set_pyomo4_expression_tree,
    detect_special_structure,
    logger,
)


class Timeout(object):
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class S3Storage(object):
    PREFIX = 's3://'

    def __init__(self):
        self.s3 = boto3.resource('s3')

    def download_to_temp(self, resource, prefix=None):
        bucket = self._bucket(resource)
        path = self._path(resource)
        temp_dir = tempfile.TemporaryDirectory(prefix=prefix)
        filename = os.path.basename(path)
        dest_file = os.path.join(temp_dir.name, filename)
        bucket.download_file(path, dest_file)
        logging.info('Downloaded {} to {}'.format(resource, dest_file))
        return dest_file, temp_dir

    def upload_file(self, orig, dest):
        bucket = self._bucket(dest)
        path = self._path(dest)
        logging.info('Uploading {} to {}'.format(orig, dest))
        bucket.upload_file(orig, path)

    def _bucket(self, resource):
        resource = resource[len(self.PREFIX):]
        parts = resource.split('/')
        return self.s3.Bucket(parts[0])

    def _path(self, resource):
        resource = resource[len(self.PREFIX):]
        parts = resource.split('/')
        return '/'.join(parts[1:])


class RunResources(object):
    """Context Manager to provide uniform interface for local and S3 objects"""
    def __init__(self, problem, solution, output):
        s3 = S3Storage()
        self.s3 = s3

        if problem.startswith(s3.PREFIX):
            # keep a ref to problem dir so it does not get cleaned self
            self.problem, self.problem_dir = \
              s3.download_to_temp(problem, prefix='problem')
        else:
            self.problem = problem

        if solution is not None and solution.startswith(s3.PREFIX):
            # keep a ref to solution dir so it does not get cleaned self
            self.solution, self.solution_dir = \
              s3.download_to_temp(solution, prefix='solution')
        else:
            self.solution = solution

        self.output_type = None
        if output is None:
            self.output_owned = False
            self.output = sys.stdout
        else:
            self.output_owned = True
            if output.startswith(s3.PREFIX):
                self.output = tempfile.NamedTemporaryFile('w', prefix='output')
                self.output_dest = output
                self.output_type = 's3'
            else:
                self.output = open(output, 'w')
                self.output_type = 'fs'

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.output_type == 'fs':
            self.output.close()
        elif self.output_type == 's3':
            self.output.flush()
            self.s3.upload_file(self.output.name, self.output_dest)


def read_problem(filename):
    if filename.endswith('.osil'):
        model = read_osil(
            filename,
            objective_prefix='obj_',
            constraint_prefix='cons_',
        )

        if model.name is None:
            basename = os.path.basename(filename)
            model.name = os.path.splitext(basename)[0]

        return model
    else:
        raise RuntimeError('Unknown file type.')


def _conscurvature(info):
    curvatures = info.conscurvature().values()
    if all(c.is_linear() for c in curvatures):
        return 'linear'
    elif all(c.is_convex() for c in curvatures):
        return 'convex'
    elif all(c.is_concave() for c in curvatures):
        return 'concave'
    else:
        return 'indefinite'


def _objsense(info):
    senses = [o['sense'] for o in info.objectives.values()]
    if len(info.objectives) == 1:
        return senses[0]
    else:
        return ', '.join(senses)


def _objcurvature(info):
    curvatures = [c for c in info.objcurvature().values()]
    if len(info.objectives) == 1:
        cvx = curvatures[0]
        if cvx.is_linear():
            return 'linear'
        elif cvx.is_convex():
            return 'convex'
        elif cvx.is_concave():
            return 'concave'
        else:
            return 'indefinite'
    else:
        return ', '.join(curvatures)


def _objtype(info):
    objtypes = info.objtype().values()
    if len(objtypes) == 1:
        return list(objtypes)[0]
    else:
        return ', '.join(objtypes)


def summarize_problem_structure(info):
    return {
        'name': info.name,
        'nvars': info.num_variables(),
        'nbinvars': info.num_binaries(),
        'nintvars': info.num_integers(),
        'ncons': info.num_constraints(),
        'conscurvature': _conscurvature(info),
        'objsense': _objsense(info),
        'objcurvature': _objcurvature(info),
        'objtype': _objtype(info),
    }


def run_for_problem(filename, solution_filename, timeout_):
    # If we can't read a problem in, don't do anything.
    logging.info('Starting {} / {}. timeout={}'.format(filename, solution_filename, timeout_))
    try:
        with Timeout(seconds=300):
            logging.info('\tReading Problem')
            model = read_problem(filename)
            logging.info('\tConverting DAG')
            model = dag_from_pyomo_model(model)
    except Exception as err:
        logging.error('{}: Error reading: {}'.format(filename, str(err)))
        return None

    try:
        with Timeout(seconds=timeout_):
            logging.info('\tStarting Special Structure Detection')
            start_t = time.time()
            info = detect_special_structure(model)
            end_t = time.time()
            logging.info('\tSpecial Structure Detection Finished')

    except Exception as err:
        logging.error('{}: {}'.format(model.name, str(err)))
        traceback.print_exc()
        return {
            'name': model.name,
            'status': 'error',
            'error': str(err),
        }

    summary = summarize_problem_structure(info)
    summary['runtime'] = end_t - start_t
    summary['status'] = 'ok'

    if solution_filename is not None:
        try:
            var_bounds_ok, obj_bounds_ok = check_solution_bounds(
                info.variables, info.objectives, solution_filename
            )
            summary['bounds_var_ok'] = var_bounds_ok
            summary['bounds_obj_ok'] = obj_bounds_ok
        except Exception as err:
            logging.error('{}: Solution bounds: {}'.format(model.name, str(err)))

    return summary


def check_solution_bounds(variables, objectives, solution_filename):
    wrong_bounds = []
    objvar_value = None
    with open(solution_filename) as f:
        for line in f:
            parts = line.split()
            var_name = parts[0]
            sol = mpf(parts[1])

            if var_name == 'objvar':
                objvar_value = sol
            if var_name not in variables:
                continue

            v = variables[var_name]

            if sol < v['lower_bound'] or sol > v['upper_bound']:
                wrong_bounds.append(var_name)

    var_bounds_ok = len(wrong_bounds) == 0
    if len(objectives) != 1 or objvar_value is None:
        return var_bounds_ok, None

    for obj in objectives.values():
        lower_bound = obj['lower_bound']
        upper_bound = obj['upper_bound']
        obj_bounds_ok = (
            objvar_value >= lower_bound and objvar_value <= upper_bound
        )
        return var_bounds_ok, obj_bounds_ok


if __name__ == '__main__':
    set_pyomo4_expression_tree()
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument('--problem', '-p')
    parser.add_argument('--solution', '-s')
    parser.add_argument('--output', '-o', nargs='?')
    parser.add_argument('--timeout', type=int, default=300,
                        help='Timeout in seconds')
    parser.add_argument('--log', dest='log_level',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        default='WARNING')
    args = parser.parse_args()

    logger.setLevel(level=args.log_level)

    if args.problem is None:
        parser.print_help()
        sys.exit(1)

    with RunResources(args.problem, args.solution, args.output) as r:
        result = run_for_problem(r.problem, r.solution, args.timeout)

        if result is None:
            sys.exit(1)

        result_str = json.dumps(result, sort_keys=True)
        r.output.write(result_str + '\n')



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = '1.2.0'
__author__ = 'Francesco Ceccon'
__author_email__ = 'francesco@ceccon.me'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018 Francesco Ceccon'
__url__ = 'https://github.com/cog-imperial/suspect'
__description__ = 'Special Structure Detection for Pyomo'
